package week6;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.File;
import java.io.IOException;

import javax.swing.JFrame;


import com.jogamp.opengl.GL2;
import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLCapabilities;
import com.jogamp.opengl.GLEventListener;
import com.jogamp.opengl.GLProfile;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.opengl.glu.GLU;
import com.jogamp.opengl.util.FPSAnimator;
import com.jogamp.opengl.util.gl2.GLUT;

import week6.Shader.CompilationException;

/**
 * Guha, Sumanta (2014-08-06). Computer Graphics Through OpenGL: From Theory to Experiments, Second Edition 
 * Experimenter code.
 *
 * Adapted to jogl and modified by Angela Finlayson 
 * Check out the keyPressed function to see what keys to press to change the settings :)
 */

/**
 * 
 * Demonstration of using shaders for Phong Per Fragment Specular lighting.
 * You can move the light with the arrows
 * Use the space bar to switch between normal fixed pipeline flat shading, garoud smooth shading  or through the shaders  
*/

public class LightAndMaterialShaders implements GLEventListener, KeyListener {
	
	private boolean light0On = true; // White light on
	private boolean light1On = false; // Blue light on
	
	//Setting for light 0
	private float a = 0.2f; // Ambient white light intensity.
	private float d = 0.5f; // Diffuse white light intensity
	private float s = 0.5f; // Specular white light intensity.
	private int p = 1; // Positional light 1, directional 0
	private float f = 80; // shininess
	private float xAngle = 0.0f, yAngle = 0.0f; // Rotation angles of white light.

	
	//Global Settings
	private float g = 0.2f; // Global Ambient intensity.
	private int localViewer = 0; // Local viewpoint?
	
	//Extra variables for different shading
	private Shading shading = Shading.FLAT; 
	private int useBlinn = 1;
	private float erosion = -0.5f;
	
	private static final String VERTEX_SHADER[] = {"src/week6/GaroudVertex.glsl",
            "src/week6/Phongvertex.glsl",
            "src/week6/Phongvertex.glsl",
            "src/week6/Phongvertex.glsl"
	};
	
	private static final String FRAGMENT_SHADER[] = {"src/week6/GaroudFragment.glsl",
            "src/week6/Phongfragment.glsl",
            "src/week6/PhongMultifragment.glsl",
            "src/week6/PhongDiscardfragment.glsl"
	};

	private int shaderprogram[] = new int[4];
	
	private Model myModel = Model.TEAPOT;
	
	public enum Shading {
		FLAT, SMOOTH, GAROUD_SHADER, PHONG_FRAGMENT, PHONG_MULTILIGHT_FRAGMENT,PHONG_DISCARD  
	};
	
	public enum Model {
		TEAPOT, CUBE, SPHERE, TORUS, BUNNY
	};

    public static void main(String[] args) {
        // initialisation
        GLProfile glp = GLProfile.getDefault();
        GLCapabilities caps = new GLCapabilities(glp);
        
        // create a panel to draw on
        GLJPanel panel = new GLJPanel(caps);

        // put it in a JFrame
        final JFrame jframe = new JFrame("Sphere In a Box");
        jframe.setSize(800, 800);
        jframe.add(panel);
        jframe.setVisible(true);

        // Catch window closing events and quit             
        jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        LightAndMaterialShaders s = new LightAndMaterialShaders();
        
        // add a GL Event listener to handle rendering
        panel.addGLEventListener(s);
        panel.addKeyListener(s);
        panel.setFocusable(true);   
        
        // NEW: add an animator to create display events at 60 FPS
        FPSAnimator animator = new FPSAnimator(60);
        animator.add(panel);
        animator.start();
    }   
    
    public void setLighting(GL2 gl) {
    	
    	// Turn lights off/on.
        if (light0On) gl.glEnable(GL2.GL_LIGHT0); else gl.glDisable(GL2.GL_LIGHT0);
        if (light1On) gl.glEnable(GL2.GL_LIGHT1); else gl.glDisable(GL2.GL_LIGHT1);
    	
        // Light property vectors.
        float lightAmb[] = { a, a, a, 1.0f };
        float lightDif0[] = { d, d, d, 1.0f };
        float lightSpec0[] = { s, s, s, 1.0f };
               
        float light1[] = { 0.0f, 0.0f, 1.0f, 1.0f };
        
        float globAmb[] = { g, g, g, 1.0f };

        // Light0 properties.
        gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_AMBIENT, lightAmb,0);
        gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_DIFFUSE, lightDif0,0);
        gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_SPECULAR, lightSpec0,0);

        // Light1 properties.
        gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_AMBIENT, light1,0);
        gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_DIFFUSE, light1,0);
        gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_SPECULAR, light1,0);

        // Global light properties
        gl.glLightModelfv(GL2.GL_LIGHT_MODEL_AMBIENT, globAmb,0); // Global ambient light.
        gl.glLightModeli(GL2.GL_LIGHT_MODEL_LOCAL_VIEWER, localViewer); // Enable local viewpoint
        drawAndPositionLights(gl);
    }
    
    private void drawAndPositionLights(GL2 gl){
    	GLUT glut = new GLUT();
    	float lightPos0[] = { 0.0f, 0.0f, 3.0f, p };
    	float lightPos1[] = { 1.0f, 2.0f, 0.0f, 1.0f };
        
    	// Light0 positioned 
    	// sphere (positional light)
        // or arrow (directional light).
        gl.glPushMatrix();{
        	//Transformations to move lights
        	gl.glRotated(xAngle, 1.0, 0.0, 0.0); // Rotation about x-axis.
        	gl.glRotated(yAngle, 0.0, 1.0, 0.0); // Rotation about z-axis.        	
            //The light pos will be subject to current transformation
        	//matrix, so will be rotated
        	gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_POSITION, lightPos0,0);
        	
        	//Also translate to draw to representation of the light
        	//Usually you would not do this if you did not
        	//want to actually draw the light.
        	gl.glTranslatef(lightPos0[0], lightPos0[1], lightPos0[2]);
        	
        	
        	float emmL[] = {1.0f, 1.0f, 1.0f, 1.0f};
        	float matAmbAndDifL[] = {0.0f, 0.0f, 0.0f, 1.0f};
        	float matSpecL[] = { 0.0f, 0.0f, 0.0f, 1.0f };
        	float matShineL[] = { f };

        	// Material properties of sphere.
        	gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, matAmbAndDifL,0);
        	gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_SPECULAR, matSpecL,0);
        	gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_SHININESS, matShineL,0);
        	
        	//Since the actual sphere/arrow representing the light will not necessarily be hit
        	//by any light and we want to, see it, we give it an emissive property.
        	//The other alternative would be to temporarily turn off lighting.
        	//We do this for the other light to give an example of both ways
        	gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_EMISSION, emmL,0);
        	if (light0On) 
        	{
        		if (p == 1) glut.glutSolidSphere(0.05, 8, 8); // Sphere at positional light source.
        		else // Arrow pointing along incoming directional light.
        		{
        			gl.glLineWidth(3.0f);
        			gl.glBegin(GL2.GL_LINES);
        			gl.glVertex3d(0.0, 0.0, 0.25);
        			gl.glVertex3d(0.0, 0.0, -0.25);
        			gl.glVertex3d(0.0, 0.0, -0.25);
        			gl.glVertex3d(0.05, 0.0, -0.2);
        			gl.glVertex3d(0.0, 0.0, -0.25);
        			gl.glVertex3d(-0.05, 0.0, -0.2);
        			gl.glEnd();
        			gl.glLineWidth(1.0f);
        		}
        	}
        }gl.glPopMatrix();

        //Light 1 positioned and drawn
        //Just for an example, instead of using emissive light we
        //Disable lighting temporarily so we can see the actual sphere
        //representing light 2
        gl.glDisable(GL2.GL_LIGHTING);
        // Light1 and its sphere positioned.
        gl.glPushMatrix();{
        	gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_POSITION, lightPos1,0);
        	gl.glTranslated(lightPos1[0], lightPos1[1], lightPos1[2]);
        	gl.glColor3f(0.0f, 0.0f, 1.0f); 
        	if (light1On) glut.glutWireSphere(0.05, 8, 8);
        }gl.glPopMatrix();
        //Enable it again for the rest of scene
        gl.glEnable(GL2.GL_LIGHTING);
    }
    
    private void setShader(GL2 gl){
	    if(shading == Shading.FLAT){
			gl.glShadeModel(GL2.GL_FLAT);
			gl.glUseProgram(0);
			System.out.println("Shading fixed function flat");
		} else if (shading == Shading.SMOOTH){
			gl.glShadeModel(GL2.GL_SMOOTH);
			gl.glUseProgram(0);
			System.out.println("Shading fixed function smooth/garoud ");
		} else {
		   int index = shading.ordinal() - 2; 
		   System.out.println("Shading " + shading);
		   if(index == 2)
		   System.out.println("Use Blinn " + useBlinn);
	       gl.glUseProgram(shaderprogram[index]);
	       gl.glUniform1f(gl.glGetUniformLocation(shaderprogram[index],"limit"),erosion);
	       gl.glUniform1i(gl.glGetUniformLocation(shaderprogram[index],"useBlinn"),useBlinn);
		}
    }
    
    @Override
    public void display(GLAutoDrawable drawable) {
    	 
        GL2 gl = drawable.getGL().getGL2(); 
        
        setShader(gl);
        
        gl.glClear (GL2.GL_COLOR_BUFFER_BIT | GL2.GL_DEPTH_BUFFER_BIT);
        gl.glMatrixMode(GL2.GL_MODELVIEW);
        gl.glLoadIdentity();

        
        
        GLUT glut = new GLUT();
        GLU glu = new GLU();
        glu.gluLookAt(0.0, 3.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
        
        setLighting(gl); 
       
        
        //Draw Model

        float matAmbAndDif[] = {1.0f, 0.0f, 0.0f, 1.0f};
        float matSpec[] = { 1.0f, 1.0f, 1.0f, 1.0f };
        float matShine[] = { f };
        float emm[] = {0.0f, 0.0f, 0.0f, 1.0f};
        
        // Material properties of teapot
        gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, matAmbAndDif,0);
        gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_SPECULAR, matSpec,0);
        gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_SHININESS, matShine,0);
        gl.glMaterialfv(GL2.GL_FRONT, GL2.GL_EMISSION, emm,0);
       
        
        
        if(myModel == Model.TEAPOT){
  		   	gl.glFrontFace(GL2.GL_CW);
  		   	glut.glutSolidTeapot(1.5);
  		   	gl.glFrontFace(GL2.GL_CCW);
  	   	} else if (myModel == Model.CUBE){
  	   		glut.glutSolidCube(2f);
  	   	} else if (myModel == Model.SPHERE){
  	   		glut.glutSolidSphere(1.5, 20, 20);
  	   	} else if (myModel == Model.TORUS){

  		   	glut.glutSolidTorus(0.5, 1.5, 20, 20);

  	   	}  else if (myModel == Model.BUNNY){
  	   		try{
  	   			gl.glPushMatrix();
  	   				gl.glScaled(3,3,3);
  	   				Bunny.gen3DObjectList(gl) ;
  	   			gl.glPopMatrix();
  		   } catch(IOException e){
  			   System.err.println("Could not open bunny.txt. Here is a donut instead");
  			   glut.glutSolidTorus(0.5, 1.5, 20, 20);
  		   }
 	   } 
        
    }

    @Override
    public void dispose(GLAutoDrawable drawable) {
    }

    @Override
    public void init(GLAutoDrawable drawable) {
    	GL2 gl = drawable.getGL().getGL2();
    
    	gl.glEnable(GL2.GL_DEPTH_TEST);
    	gl.glEnable(GL2.GL_LIGHTING);
    	// Cull back faces.
    	//gl.glEnable(GL2.GL_CULL_FACE);
    	//gl.glCullFace(GL2.GL_BACK);
    	
    	gl.glEnable(GL2.GL_NORMALIZE);
 	   
    	// init shaders
        try {
            initShaders(gl);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    @Override
    public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h) {
    	 GL2 gl = drawable.getGL().getGL2();
         gl.glMatrixMode(GL2.GL_PROJECTION);
         gl.glLoadIdentity();
         
         GLU glu = new GLU();
         glu.gluPerspective(60.0, (float)w/(float)h, 1.0, 20.0);
         
         gl.glMatrixMode(GL2.GL_MODELVIEW);
    }

    @Override
	public void keyPressed(KeyEvent ev) {
    	switch (ev.getKeyCode()) {
    	case KeyEvent.VK_W:			
    		light0On = !light0On;

    		break;
    	case KeyEvent.VK_B:
    		light1On = !light1On;
    		break;
    	case KeyEvent.VK_R:
    		if(useBlinn == 0) useBlinn = 1;
    		else useBlinn = 0;

    		break;
    	case KeyEvent.VK_L:
    		if(localViewer == 1) localViewer = 0;
    		else localViewer = 1;
    		System.out.println("Local viewer " + localViewer);
    		break;
    	case KeyEvent.VK_P:
    		//Positional vs directional
    		if(p == 1) p = 0;
    		else p = 1;

    		break;
    	case KeyEvent.VK_D:
    		if (ev.isShiftDown()) {
    			if (d < 1.0) d += 0.05;
    		} else {
    			if (d > 0.0) d -= 0.05;
    		}

    		break;
    	case KeyEvent.VK_F:
    		if (ev.isShiftDown()) {
    			if (f < 128.0) f += 1;
    		} else {
    			if (f > 1.0) f -= 1;
    		}

    		break;
    	case KeyEvent.VK_A:
    		if (ev.isShiftDown()) {
    			if (a < 1.0) a += 0.05;
    		} else {
    			if (a > 0.0) a -= 0.05;
    		}
    		break;	
    	case KeyEvent.VK_S:
    		if (ev.isShiftDown()) {
    			if (s < 1.0) s += 0.05;
    		} else {
    			if (s > 0.0) s -= 0.05;
    		}
    		break;	
    	case KeyEvent.VK_G:
    		if (ev.isShiftDown()) {
    			if (g < 1.0) g += 0.05;
    		} else {
    			if (g > 0.0) g -= 0.05;
    		}
    		break;	
    	case KeyEvent.VK_UP:
    		xAngle--;
    		if (xAngle < 0.0) xAngle += 360.0;
    		break;
    	case KeyEvent.VK_DOWN:
    		xAngle++;
    		if (xAngle > 360.0) xAngle -= 360.0;
    		break;
    	case KeyEvent.VK_LEFT:
    		yAngle--;
    		if (yAngle < 0.0) yAngle += 360.0;
    		break;
    	case KeyEvent.VK_RIGHT:
    		yAngle++;
    		if (yAngle > 360.0) yAngle -= 360.0;
    		break;
    	case KeyEvent.VK_E:
			 if (ev.isShiftDown()) {
			     erosion-= 0.01;	 
			 } else {
	    		erosion+= 0.01;
			 }		
			 System.out.println(erosion);
	         break;
    	case KeyEvent.VK_SPACE:
    		 if (ev.isShiftDown()) {
			     shading = Shading.values()[(shading.ordinal() - 1)%6]; 
    		 } else {
    			 shading = Shading.values()[(shading.ordinal() + 1)%6]; 
			 }
			 break;
		  case KeyEvent.VK_1:
	            myModel = Model.TEAPOT;
	            break;
	            
		  case KeyEvent.VK_2:
			  myModel = Model.CUBE;
			  break;

		  case KeyEvent.VK_3:
			  myModel = Model.SPHERE;
			  break;

		  case KeyEvent.VK_4:
			  myModel = Model.TORUS;
			  break;
			  
		  case KeyEvent.VK_5:
			  myModel = Model.BUNNY;
			  break;
    	default:
    		break;
    	}
    }

	@Override
	public void keyReleased(KeyEvent arg0) {
		
	}

	@Override
	public void keyTyped(KeyEvent arg0) {
		
	}
	
	 private void initShaders(GL2 gl) throws Exception {
	        for(int i=0; i < shaderprogram.length; i++){
	        	
	        
	    	     shaderprogram[i] = initShader(gl,i);
	        }
	    }
	    private int initShader(GL2 gl, int index) throws Exception {
	        Shader vertexShader = new Shader(GL2.GL_VERTEX_SHADER, new File(VERTEX_SHADER[index]));
	        vertexShader.compile(gl);
	        Shader fragmentShader = new Shader(GL2.GL_FRAGMENT_SHADER, new File(FRAGMENT_SHADER[index]));
	        fragmentShader.compile(gl);
	        
	        //Each shaderProgram must have
	        //one vertex shader and one fragment shader.
	        int shaderprogram = gl.glCreateProgram();
	        gl.glAttachShader(shaderprogram, vertexShader.getID());
	        gl.glAttachShader(shaderprogram, fragmentShader.getID());
	        
	        
	        gl.glLinkProgram(shaderprogram);
	        
	        
	        int[] error = new int[2];
	        gl.glGetProgramiv(shaderprogram, GL2.GL_LINK_STATUS, error, 0);
	        if (error[0] != GL2.GL_TRUE) {
	        	int[] logLength = new int[1];
	            gl.glGetProgramiv(shaderprogram, GL2.GL_INFO_LOG_LENGTH, logLength, 0);

	            byte[] log = new byte[logLength[0]];
	            gl.glGetProgramInfoLog(shaderprogram, logLength[0], (int[]) null, 0, log, 0);

	            System.out.printf("Failed to compile shader! %s\n", new String(log));
	            throw new CompilationException("Error compiling the shader: "
	                    + new String(log));
	        }
	        
	        gl.glValidateProgram(shaderprogram);
	        
	        gl.glGetProgramiv(shaderprogram, GL2.GL_VALIDATE_STATUS, error, 0);
	        if (error[0] != GL2.GL_TRUE) {
	        	System.out.printf("Failed to validate shader!\n");
	        	throw new Exception("program failed to validate");
	        }

	        return shaderprogram;

	    }
}
